#! /usr/bin/env python3
import sys
import os
import csv
from random import seed, random
from pprint import pprint as pp
from math import log, exp
import numpy as np

removed=[]

def doNormalize(X,isTrain):
    #do 0 mean 1 std normalization
    x1 = np.array(X,dtype=float)
    for i in range(len(X[0])):
        col = x1[:,i]
        mean,std = col.mean(),col.std()
        #std = std if std!=0.0 else 1.0
        if i!=0:
            if std<0.1 and isTrain:
                removed.append(i)
            else:
                x1[:,i] = (x1[:,i]-mean)/max(std,1.0)
    x1 = np.delete(x1, removed, axis=1)
    return x1.tolist()

def getData(srcF, isTrain=True, addBias=True, normalize=True):
    X,y=[],[]
    with open(srcF) as src:
        reader = csv.reader(src, delimiter=',')
        for i,row in enumerate(reader):
            temp = []
            if addBias:
                temp.append(1)
            end = -1 if isTrain else len(row)
            temp.extend(row[:end])
            
            #correct data type
            X.append(list(map(float, temp)))

            if isTrain:
                v = int(row[-1])
                entry = [1,0] if v==1 else [0,1]
                y.append(entry)

    if normalize:
        X = doNormalize(X,isTrain)
    #print(X[0])
    return (np.array(X),np.array(y))

def sigmoid(v):
    z = np.minimum(-v, 700.0)
    return 1/(1+np.exp(z))

def sigmoidDiff(v):
    return sigmoid(v)*(1-sigmoid(v))

def feedforward(model, X):
    w1,w2 = model['w1'],model['w2'] #58x28, 29x2
    z1 = X.dot(w1)  #mx58 * 58x28 = mx28  
    a1 = sigmoid(z1)    #mx28
    a1 = np.insert(a1,0,np.ones(a1.shape[0]),axis=1)  #mx29
    z2 = a1.dot(w2) #mx29 * 29x2
    h = sigmoid(z2) #mx2    
    return h

def cost(model, X, y):    
    m = X.shape[0]
    h = feedforward(model, X)

    h = np.minimum(np.maximum(h, 1e-15),1-1e-15)
    
    # vf = np.vectorize(restrictProb)
    # py = vf(h)

    loss = -(1.0/m)*np.sum(y*np.log(h)+(1-y)*np.log(1-h))  #mx2 .* mx2
    #regularize
    w1,w2 = model['w1'],model['w2']
    loss += model['lambda']*(np.sum(np.square(w1)) + np.sum(np.square(w2)))/(2*m)
    return loss

def predict(model, x):
    w1,w2 = model['w1'],model['w2']
    #print(x.shape, w1.shape)
    z1 = x.dot(w1)  #1x58 * 58x28 = 1x28
    a1 = sigmoid(z1)    #1x28
    a1 = np.insert(a1,0,1)
    z2 = a1.dot(w2)  #1x29 x 29x2
    h = sigmoid(z2)
    return 1-np.argmax(h)


def fit(model, X, y, passes=1000):
    m = X.shape[0]
    w1,w2 = model['w1'],model['w2'] #58x28, 29x2
    li,lh,lo=model['li'],model['lh'],model['lo']
    
    for i in range(passes):
        z1 = X.dot(w1)  #mx58 * 58x28 = mx28
        a2 = sigmoid(z1)    #mx28
        a2 = np.insert(a2,0,np.ones(a2.shape[0]),axis=1)  #mx29
        z2 = a2.dot(w2) #mx29 * 29x2
        h = sigmoid(z2) #mx2

        #backpropagation
        del3 = h-y  #mx2
        z1 = np.insert(z1,0,np.ones(z1.shape[0]),axis=1)
        #mx29
        del2 = del3.dot(w2.reshape(lo,lh+1))*sigmoidDiff(z1)
        #mx2 * 2x29 .* mx29 = mx29
        del2 = del2[:,1:]   #mx28
        dw1 = np.dot(X.T,del2)    #58xm*mx28=58x28
        dw2 = (a2.T).dot(del3)    #29xm*mx2=29x2

        dw1 += (model['lambda']/m)*w1
        dw2 += (model['lambda']/m)*w2

        costs = []
        ws = []
        for eta in model['eta']:
            tw1 = w1-eta*dw1
            tw2 = w2-eta*dw2
            model['w1'] = tw1
            model['w2'] = tw2
            costs.append(cost(model,X,y))
            ws.append((tw1,tw2))
        idx = np.argmin(costs)
        w1,w2 = ws[idx][0],ws[idx][1]
        model['w1'],model['w2'] = w1,w2

        if i % (passes/10)==0:
            print(i,costs[idx])

    return model


if __name__ == "__main__":
    np.random.seed(47)
    np.seterr(over='raise')

    X,y = getData("Train.csv")
    tX,ty = getData("TestX.csv",isTrain=False)

    model = {}
    model = {'li':X.shape[1]-1,'lh':int(33*(X.shape[1]-1)/20),'lo':2,'lambda':0.1,'eta':[0.01,0.03,0.06,0.1,0.2]}
    # model['w1'] = np.random.randn(model['li']+1, model['lh'])/np.sqrt(model['li']+1)  #58x28
    # model['w2'] = np.random.randn(model['lh']+1, model['lo'])/np.sqrt(model['lh']+1)    #29x2
    model['w1'] = np.random.rand(model['li']+1, model['lh'])*0.24 - 0.12
    model['w2'] = np.random.rand(model['lh']+1, model['lo'])*0.24 - 0.12
    
    #cost(model, X, y)
    # for h in [57/3, 57/2, 2*57/3, 57, 3*57/2]:
    #     h=int(h)
    #     model = {'li':57,'lh':h,'lo':2,'lambda':0.1,'eta':0.1}    
    #     model['w1'] = np.random.randn(model['li']+1, model['lh'])/np.sqrt(model['li']+1)  #58x28
    #     model['w2'] = np.random.randn(model['lh']+1, model['lo'])/np.sqrt(model['lh']+1)    #29x2
        
    model = fit(model, X, y, passes=1500)

    m = X.shape[0]
    py,y2=[],[]
    for i,row in enumerate(tX):
        ans = predict(model, np.array(row))
        py.append(ans)
        # y2.append(1 if y[i][0]==1 else 0)

    #     acc = m-np.sum(abs(np.array(py)-np.array(y2)))
    #     print(h, acc*100/m)

    with open("answer.txt","w") as wfile:
        writer = csv.writer(wfile)
        writer.writerow(['Id','Label'])
        for i,ans in enumerate(py):
            writer.writerow([i,ans])

    # acc = m-np.sum(abs(np.array(py)-np.array(y2)))
    # print(acc*100/m)
